import io
import numpy as np
import re
from functools import wraps
from textwrap import dedent

from neatlogger import log


class SafeDict(dict):
    """
    A variant of dict that does not raise an error for missing keys.
    Thanks to: https://stackoverflow.com/a/17215533/2575273

    Example
    -------
    >>> parameters = SafeDict(a=1, b=2)
    ... print(parameters["a"], parameters["c"])
    1 {c}

    """

    def __missing__(self, key):
        return "{" + key + "}"


class Magazine:
    """
    Can be used to log information in a human-readable way.
    Supports different report topics.
    The list of topics can be later posted as a composite string.
    Useful for writing reports with class Publish().

    Examples
    --------
    >>> Magazine.report("observations", "Temperature today was {:.2f}.", None)
    ... Magazine.report("observations", "Data was corrected following {:}, only {:d} points remained.", "Brown et al. (1979)", 42)
    ... Magazine.post("observations")
    Temperature today was nan. Data was corrected following Brown et al. (1979), only 42 points remained.

    """

    topics = dict()
    figures = dict()
    references = []
    dois = []

    def __init__(self):
        pass

    @staticmethod
    def assert_topic(topic: str):
        """
        Makes sure that the topic exists in dict before appending.
        Intialized as empty list per topic.

        Parameters
        ----------
        topic: str
            Name of an existing or new topic

        Examples
        --------
        >>> Magazine.assert_topic("Experiments")
        ... Magazine.topics["Experiments"]
        []

        """
        if not topic in Magazine.topics:
            Magazine.topics[topic] = []
            Magazine.figures[topic] = []

    @staticmethod
    def report(topic="default", message="", *values):
        """
        Appends a text or image to the topic's list.
        The text is checked for Nonetype values before.

        Parameters
        ----------
        topic: str
            Name of an existing or new topic
        message: str | io.BytesIO
            Text or bytes object (to store figures)
        *values
            Any number of values to be inserted into the formatted message

        Examples
        --------
        >>> Magazine.report("Experiments", "Today is {}.", "Monday")

        """
        Magazine.assert_topic(topic)

        if isinstance(message, str):
            # normal text
            if values:
                # Replace all None by np.nan to avoid NoneType Error on formatting
                values = [np.nan if v is None else v for v in values]
                message = message.format(*values)

            Magazine.topics[topic].append(message)

        elif isinstance(message, io.BytesIO):
            # figure object
            Magazine.figures[topic].append(message)

        else:
            log.warning("Nothing to report: message is neither text nor image.")

    @staticmethod
    def cite(*refs):
        """
        Appends a reference to the Magazine that can be later converted to a reference list.

        Parameters
        ----------
        *refs: str
            Any number of reference texts.
            If it looks like a doi, it will be stored in a separate list which is used to download full texts by collect_references().

        Examples
        --------
        >>> Magazine.cite("10.5194/hess-27-723-2023", "10.1029/2021gl093924")
        >>> Magazine.cite("Einstein, A. (1916). Die Grundlage der allgemeinen Relativitätstheorie. Annalen Der Physik, 354(7), 769–822. Portico")

        """
        doi_pattern = r"^10[.][0-9]{4,}"

        for ref in refs:
            if re.search(doi_pattern, ref):
                Magazine.dois.append(ref)
            else:
                Magazine.references.append(ref)

    @staticmethod
    def post(*topics) -> str:
        """
        Joins the topic's list on a single space.

        Parameters
        ----------
        *topics: str
            Any number of existing topics

        Returns
        -------
        str
            Merged topic texts.

        Examples
        --------
        >>> paragraph = Magazine.post("Experiments", "Methods")
        """
        # if isinstance(topic, str):
        #     topic = [ topic ]
        text = []
        for topic in topics:
            Magazine.assert_topic(topic)
            text.append(" ".join(Magazine.topics[topic]))

        return " ".join(text)

    @staticmethod
    def figure(*topics) -> list:
        """
        Joins the topic's figures to a combined flat list.

        Parameters
        ----------
        *topics: str
            Any number of existing topics

        Returns
        -------
        list
            Merged topic figures.

        Examples
        --------
        >>> all_figures = Magazine.figure("Experiments", "Methods")

        """
        # if isinstance(topic, str):
        #     topic = [ topic ]
        figures = []
        for topic in topics:
            Magazine.assert_topic(topic)
            for figure in Magazine.figures[topic]:
                figures.append(figure)

        return figures

    @staticmethod
    def collect_references() -> list:
        """
        Lists all items in Magazine.references.
        Downloads the full reference text for all items in Magazine.dois

        Returns
        -------
        list
            List of reference texts, sorted by name.

        Examples
        --------
        >>> for item in Magazine.collect_references():
        ...     print(item)

        """
        reflist = Magazine.references

        if len(Magazine.dois) > 0:

            Magazine.dois = list(set(Magazine.dois))  # remove duplicates
            log.progress("Collecting {} citations from CrossRef...", len(Magazine.dois))
            from habanero import cn

            reflist2 = cn.content_negotiation(ids=Magazine.dois, format="text")
            if isinstance(reflist2, str):
                reflist2 = [reflist2]
            reflist2 = [ref.rstrip() for ref in reflist2 if ref is not None]
            reflist += reflist2

        reflist.sort()
        return reflist

    @staticmethod
    def clean():
        """
        Cleans topics, figures, references, and dois to make space for a new Magazine.
        """
        Magazine.topics = dict()
        Magazine.figures = dict()
        Magazine.references = []
        Magazine.dois = []
        return

    new = clean

    class reporting:
        def __init__(self, topic):
            """
            Decorator to report input, output, and other exposed variables.

            Parameters
            ----------
            topic (str): The topic of the function.

            """
            self.topic = topic
            self.parameters = SafeDict()

        def __call__(self, func):
            @wraps(func)
            def wrapper(*args, **kwargs):

                # Add the helper to the function's local namespace
                func_globals = func.__globals__
                func_globals["_report"] = self.parameters

                # Call the actual function and clean up afterwards
                try:
                    result = func(*args, **kwargs)
                finally:
                    del func_globals["_report"]

                # Set some predefined parameters
                self.parameters["function"] = func.__name__
                self.parameters["args"] = args
                self.parameters["return"] = result
                # Set input parameters
                for kw in kwargs:
                    self.parameters[kw] = kwargs[kw]

                # Parse the docstring
                self._parse_docstring(func.__doc__)

                return result

            return wrapper

        def _parse_docstring(self, docstring):
            """Parses the docstring for Report and References sections."""

            if docstring:
                # Report
                # ------
                report_pattern = r"Report\s+\-+\n(.*?)(?:\n\n|\Z)"
                report_match = re.search(report_pattern, docstring, re.DOTALL)
                if report_match:
                    report_text = report_match.group(1)
                    report_text = report_text.format_map(self.parameters)
                    report_text = dedent(report_text)
                    # remove single newlines
                    report_text = re.sub(r"(?<!\n)\n(?!\n)", " ", report_text)
                    Magazine.report(self.topic, report_text)

                # References
                # ----------
                refs_pattern = r"References\s+\-+\n(.*?)(?:\n\n|\Z)"
                refs_match = re.search(refs_pattern, docstring, re.DOTALL)
                if refs_match:
                    refs_text = refs_match.group(1)
                    refs_text = dedent(refs_text)
                    # Create list from lines
                    refs_list = refs_text.split("\n")
                    Magazine.cite(*refs_list)

            else:
                log.warning(
                    "No docstring provided for function {}.",
                    self.parameters["function"],
                )



import sys
import os
from os.path import dirname, realpath

from neatlogger import log


def get_file_size(file, human_readable=False):
    try:
        size = os.stat(file).st_size
    except Exception as e:
        log.error("Cannot retrieve size of file {}: {}", file, e)

    if human_readable:
        if size < 1024:
            sizestr = "%.0f Bytes" % (size)
        elif size < 1024**2:
            sizestr = "%.0f KB" % (size / 1024)
        else:
            sizestr = "%.1f MB" % (size / 1024 / 1024)
        return sizestr
    else:
        return size


def assert_directory(path):

    (folder, filename) = os.path.split(path)
    if folder != "" and not os.path.isdir(folder):
        try:
            os.makedirs(folder)
            log.success("Folder created: {}", folder)
        except Exception as e:
            log.error("Folders {} cannot be created: {}", folder, e)


@log.catch
def get_script_directory():
    if getattr(sys, "frozen", False):
        # application_path = abspath(dirname(dirname(realpath(sys.executable))))
        path_of_script = dirname(realpath(sys.executable))  # sys._MEIPASS
    elif __file__:
        path_of_script = dirname(__file__)
        # application_path = abspath(dirname(dirname(__file__)))
    # log.info("Script path: {}", path_of_script)
    return path_of_script



from datetime import datetime
import pandas as pd
from neatlogger import log

from magazine import Magazine
from magazine.io import get_file_size, assert_directory, get_script_directory


# Requires FPDF, https://py-pdf.github.io/fpdf2/
import fpdf

# log.getLogger("fpdf.svg").propagate = False


class Publish:
    """
    Context manager to write all reports, figures, and references into a PDF file.
    Uses FPDF2 to return a PDF class.

    Parameters
    ----------
    filename : str
        Full path and name of the PDF file.
    title : str, optional
        Title to be written in the header, by default ""
    info : str, optional
        Any info to be written in the header, e.g. a version number, by default ""
    datetime_fmt : str, optional
        How to format the datetime in the header, by default "%Y-%m-%d %H:%M"
    page_numbers : bool, optional
        Whether or not to show page numbers, by default True

    Returns
    -------
    PDF
        An instance that inherits from fpdf.FPDF to provide commands to add content to the PDF.

    Examples
    --------
    >>> with Magazine.Publish("example.pdf", "My Title") as M:
    ...     M.add_page()
    ...     M.add_title("Chapter 1)
    ...     M.add_paragraph("Long text")
    ...     M.add_image(figure)
    ...     M.add_table(data)

    """

    def __init__(
        self,
        filename: str,
        title: str = "",
        info: str = "",
        datetime_fmt: str = "%Y-%m-%d %H:%M",
        page_numbers: bool = True,
    ):
        self.filename = filename
        self.file_format = filename[filename.rindex(".") + 1 :].lower()
        self.title = title
        self.info = info
        self.page_numbers = page_numbers
        self.datetime_fmt = datetime_fmt
        self.magazine = None

    def __enter__(self):
        if self.file_format == "pdf":
            self.magazine = PDF(
                self.title, self.info, self.datetime_fmt, self.page_numbers
            )
            return self.magazine
        else:
            log.error(
                "The requested magazine format is not supported: {}", self.file_format
            )
            return None

    def __exit__(self, type, value, traceback):
        if self.magazine is None:
            pass
        else:
            assert_directory(self.filename)
            self.magazine.output(self.filename)
            log.success(
                "Magazine published: {} ({})",
                self.filename,
                get_file_size(self.filename, human_readable=True),
            )


######################
class PDF(fpdf.FPDF):
    """
    PDF writer and wrapper for FPDF

    Examples
    --------
    >>> pdf = PDF()
    ... pdf.header_text = "My title"
    ... pdf.add_page()
    ... pdf.output("my_title.pdf")

    Notes
    -----
    Thanks to: https://py-pdf.github.io/fpdf2/Maths.html

    """

    # These variables can be changed individually if necessary
    cell_height = 8
    header_text = ""
    font = "Roboto"
    font_mono = "RobotoM"  # "Courier" # Helvetica
    font_size = 10
    ln0 = dict(new_x=fpdf.enums.XPos.RIGHT, new_y=fpdf.enums.YPos.TOP)
    ln1 = dict(new_x=fpdf.enums.XPos.LMARGIN, new_y=fpdf.enums.YPos.NEXT)

    def __init__(
        self,
        title: str = "",
        info: str = "",
        datetime_fmt: str = "",
        page_numbers: bool = True,
    ):
        super().__init__()

        self.title = title
        self.info = info
        self.datetime_fmt = datetime_fmt
        self.page_numbers = page_numbers

        self.header_text = self.title

        # fonts
        font_folder = get_script_directory() + "/fonts/"
        # font_folder = os.path.dirname(os.path.abspath(__file__)) + "/../app/_ui/fonts/"
        self.add_font("Roboto", "", font_folder + "Roboto-Regular.ttf")
        self.add_font("Roboto", "B", font_folder + "Roboto-Bold.ttf")
        self.add_font("RobotoM", "", font_folder + "RobotoMono-Regular.ttf")
        self.add_font("RobotoM", "B", font_folder + "RobotoMono-Bold.ttf")

    def header(self):
        """
        Overwrites the FPDF's header function with a table:
        | %title | %info | %datetime | %page |
        """

        # title
        self.set_font(self.font, "B", self.font_size)
        self.cell(
            self.epw - 35 - 45 - 15,
            8,
            " %s" % self.header_text,
            border=True,
            align="L",
            **self.ln0
        )

        # info
        self.set_font(self.font_mono, "", self.font_size)
        self.cell(35, 8, self.info, border=True, align="C", **self.ln0)

        # datetime
        datetime_str = (
            "" if not self.datetime_fmt else datetime.now().strftime(self.datetime_fmt)
        )
        self.cell(45, 8, datetime_str, border=True, align="C", **self.ln0)

        # page
        page_str = "" if not self.page_numbers else "%2s " % str(self.page_no())
        self.cell(15, 8, page_str, border=True, align="R", **self.ln1)
        self.ln(self.cell_height)

    # def footer(self):
    #     self.set_y(-15)
    #     self.set_font('Courier', '', 12)
    #     self.cell(0, 8, f'Page {self.page_no()}', True, align='C', **ln0)

    def add_title(self, title: str = None, style: str = "B"):
        """
        Add a chapter title.

        Parameters
        ----------
        title: str
            Title of the page, can be empty.
        style: str, optional
            Font style, is "B" for bold by default.

        Examples
        --------
        >>> with Publish("example.pdf") as M:
        ...     M.add_title("My title")

        """
        if title is None:
            title = self.header_text
        self.set_font(self.font, style=style, size=24)
        self.cell(w=0, h=20, text=title, **self.ln1)
        self.set_font(style="", size=self.font_size)
        # self.ln(self.cell_height)

    add_head = add_title

    def add_paragraph(self, text: str = None):
        """
        Add a multiline paragraph.

        Parameters
        ----------
        text: str
            Text to be written.

        Examples
        --------
        >>> with Publish("example.pdf") as M:
        ...     M.add_paragraph("Very long text.")

        """
        if text is None:
            text = ""
        self.multi_cell(w=0, h=5, text=text)
        self.ln(self.cell_height)

    add_text = add_paragraph

    def add_topic(self, topic: str = None, headers: bool = True, new_page: bool = True):
        """
        Shortcut to add a page, title, and topic text.

        Parameters
        ----------
        topic : str
            Topic to take the report from.
        headers : bool, optional
            Write a headline, by default True
        new_page : bool, optional
            Create a new page, by default True

        Examples
        --------
        >>> with Publish("example.pdf") as M:
        ...     M.add_topic("Experiments")
        """
        if new_page:
            self.add_page()
        if headers:
            self.add_title(topic)
        self.add_paragraph(Magazine.post(topic))

    def add_image(
        self,
        source=None,
        x: float = None,
        y: float = None,
        w: float = None,
        h: float = 0,
        link: str = "",
    ):
        """
        Write an image to PDF.

        Parameters
        ----------
        source : string or object or list of objects, optional
            Can be a file path (png, jpg) or an image buffer or a list of them., by default None
        x : float, optional
            x coords of image, by default None
        y : float, optional
            y coords of image, by default None
        w : float, optional
            width of image, by default None
        h : float, optional
            height of image, scales automatically with width, by default 0
        link : str, optional
            link for the image, by default ""

        Examples
        --------
        >>> image_object = io.BytesIO()
        ... plt.savefig(image_object, format="svg")
        ... with Publish("example.pdf") as M:
        ...     M.add_image(image_object)
        ...     M.add_image("image_file.png")

        """

        if w is None:
            w = self.epw

        if not isinstance(source, list):
            source = [source]

        for obj in source:
            if obj:
                self.image(obj, x=x, y=y, w=w, h=h, link=link)
                self.ln(self.cell_height)

    def add_figure(
        self, topic: str = None, headers: bool = False, new_page: bool = False
    ):
        """
        Write all figures of a topic to the PDF.

        Parameters
        ----------
        topic : str, optional
            Existing topic, by default None
        headers : bool, optional
            Add a topic headline before the figure, by default False
        new_page : bool, optional
            Create a new page, by default False

        Examples
        --------
        >>> image_object = io.BytesIO()
        ... plt.savefig(image_object, format="svg")
        ... Magazine.report("Experiments", image_object)
        ... with Publish("example.pdf") as M:
        ...     M.add_figure("Experiments")

        """
        if new_page:
            self.add_page()
        if headers:
            self.add_title(topic)
        self.add_image(Magazine.figure(topic))

    def add_table(
        self, data: pd.DataFrame = None, align: str = "RIGHT", index: bool = False
    ):
        """
        Add a table for a pandas DataFrame.

        Parameters
        ----------
        data : pd.DataFrame, optional
            DataFrame to print to the PDF, by default None
        align : str, optional
            Alignment of the cell text, by default "RIGHT"
        index : bool, optional
            Whether or not to display the index column, by default False

        Examples
        --------
        >>> data = pd.DataFrame({"name":["Tim", "Tom"], "age":[12,13]})
        ... with Publish("example.pdf") as M:
        ...     M.add_table(data, index=True)

        """
        self.set_font(self.font_mono, size=7)

        if "Date" in data.columns:
            data["Date"] = data.index.strftime("%Y-%m-%d")
        if index:
            data[data.index.name] = data.index

        cols = list(data.columns)
        cols = [cols[-1]] + cols[:-1]
        data = data[cols]

        data = data.astype(str)
        columns = [list(data)]  # Get list of dataframe columns
        rows = data.values.tolist()  # Get list of dataframe rows
        data = columns + rows  # Combine columns and rows in one list

        with self.table(
            borders_layout="SINGLE_TOP_LINE",
            cell_fill_color=245,
            cell_fill_mode="ROWS",
            line_height=self.font_size * 0.5,
            text_align=align,
            width=self.epw,
        ) as table:
            for data_row in data:
                row = table.row()
                for datum in data_row:
                    row.cell(datum)

        self.set_font(self.font, style="", size=self.font_size)
        self.ln(self.cell_height)

    def add_references(self, headers: str = "References", new_page: bool = True):
        """
        Create a list of references that were previously added by Magazine.cite()
        This function will look up the full citations text using the habanero package.

        Parameters
        ----------
        headers : str, optional
            Title of the references page, by default "References"
        new_page : bool, optional
            Whether or not to add a new page, by default True

        Examples
        --------
        >>> Magazine.cite("10.1029/2021gl093924")
        ... with Publish("example.pdf") as M:
        ...     M.add_references()

        """

        if new_page:
            self.add_page()
        if headers:
            self.add_title("References")

        reftexts = Magazine.collect_references()

        # for ref in reftexts:
        #   self.add_paragraph(ref)
        self.add_paragraph("\n\n".join(reftexts))
